#! /usr/bin/env python
#
# Unit tests for the notebook SPARCL_client_method_tests.ipynb
#
# Created by nb2test.py from SPARCL_client_method_tests.ipynb
#
# To run tests:
#   python -m unittest methods_tests.py
#
###############################################################################

# START Pre-notebook code

import unittest

#! from unittest import skip

# END Pre-notebook code

###########################################################################
# Notebook combined into one function per Section
#


# NoSection
__author__ = "Alice Jacques <alice.jacques@noirlab.edu>"
__version__ = "20230608"  # yyyymmdd;
__datasets__ = ["sdss_dr16", "boss_dr16", "desi_edr"]
#!pip install sparclclient==1.2.0b4
# SPARCL imports
from sparcl.client import SparclClient

# 3rd party imports
#! from time import time

# Data Lab imports
# from dl import queryClient as qc
# from dl import authClient as ac
# from getpass import getpass
client = SparclClient(url="https://sparc1.datalab.noirlab.edu")
client


def get_all_fields():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss
    global outs, some_missing_ids, some_missing_specids, specids_any
    global specids_boss, specids_desi, specids_lots, specids_sdss
    #! client.get_all_fields?
    # Get all fields common to all datasets
    print(client.get_all_fields())


def get_all_fields_dataset_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss
    # Get all fields in SDSS-DR16
    print(client.get_all_fields(dataset_list=["SDSS-DR16"]))
    # Get all fields in BOSS-DR16
    print(client.get_all_fields(dataset_list=["BOSS-DR16"]))
    # Get all fields in DESI-EDR
    print(client.get_all_fields(dataset_list=["DESI-EDR"]))
    # Get all fields common to both SDSS-DR16 and DESI-EDR
    print(client.get_all_fields(dataset_list=["SDSS-DR16", "DESI-EDR"]))


def get_default_fields():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss
    #! client.get_default_fields?
    # Get default fields common to all datasets
    print(client.get_default_fields())


def get_default_fields_dataset_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss
    # Get default fields in SDSS-DR16
    print(client.get_default_fields(dataset_list=["SDSS-DR16"]))
    # Get default fields in BOSS-DR16
    print(client.get_default_fields(dataset_list=["BOSS-DR16"]))
    # Get default fields in DESI-EDR
    print(client.get_default_fields(dataset_list=["DESI-EDR"]))
    # Get default fields common to both SDSS-DR16 and DESI-EDR
    print(client.get_default_fields(dataset_list=["SDSS-DR16", "DESI-EDR"]))


def get_available_fields():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    #! client.get_available_fields?
    # Get available fields common to all datasets
    print(client.get_available_fields())


def get_available_fields_dataset_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    # Get available fields in SDSS-DR16
    print(client.get_available_fields(dataset_list=["SDSS-DR16"]))
    # Get available fields in BOSS-DR16
    print(client.get_available_fields(dataset_list=["BOSS-DR16"]))
    # Get available fields in DESI-EDR
    print(client.get_available_fields(dataset_list=["DESI-EDR"]))
    # Get available fields common to both SDSS-DR16 and DESI-EDR
    print(client.get_available_fields(dataset_list=["SDSS-DR16", "DESI-EDR"]))


def find():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss
    #! client.find?
    outs = [
        "data_release",
        "datasetgroup",
        "dateobs_center",
        "dec",
        "exptime",
        "sparcl_id",
        "instrument",
        "ra",
        "redshift",
        "redshift_err",
        "redshift_warning",
        "site",
        "specid",
        "specprimary",
        "spectype",
        "targetid",
        "telescope",
        "wavemax",
        "wavemin",
    ]


def find_limit():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss
    #! # does the default limit work?
    client.find()
    #! # does setting a limit less than the default value work?
    client.find(limit=2)
    # does setting a limit over the default value but under
    #! # the maximum value work?
    client.find(limit=800)
    #! # does setting the limit to None work?
    client.find(limit=None)


def find_constraints():  # noqa: C901
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss
    # No constraints passed

    found = client.find(outfields=outs, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # No outfields passed

    cons = {
        "data_release": ["SDSS-DR16"],
        "dec": [4, 29],
        "exptime": [2000, 12000],
        "ra": [80, 136],
        "redshift": [2, 5],
    }

    found = client.find(constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"First record: {found.records[0]}")
    # Invalid constraint value: data_release

    cons = {"data_release": [{"poppy"}]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: datasetgroup

    cons = {"datasetgroup": [{"daffodil"}]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: dateobs_center

    cons = {"dateobs_center": ["2011-12-21 12:00"]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: dec

    cons = {"dec": [29]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: exptime

    cons = {"exptime": [30]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: instrument

    cons = {"instrument": [{"hyacinth"}]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: ra

    cons = {"ra": [80]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: redshift

    cons = {"redshift": [1]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: redshift_err

    cons = {"redshift_err": [0.0001]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: redshift_warning

    cons = {"redshift_warning": ["magnolia"]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: site

    cons = {"site": [{"lotus"}]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: specprimary

    cons = {"specprimary": ["chrysanthemum"]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: spectype

    cons = {"spectype": [{"tulip"}]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: telescope

    cons = {"telescope": [{"carnation"}]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: wavemax

    cons = {"wavemax": [10500]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Invalid constraint value: wavemin

    cons = {"wavemin": [3800]}

    try:
        found = client.find(outfields=outs, constraints=cons, limit=10)
    except Exception as e:
        print(e)
    # Test all possible constraint fields with specified constraint values,
    # SDSS-DR16 only

    cons = {
        "data_release": ["SDSS-DR16"],
        "datasetgroup": ["SDSS_BOSS"],
        "dateobs_center": ["2003-11-21 00:00", "2011-12-21 00:00"],
        "dec": [4, 29],
        "exptime": [2000, 12000],
        "instrument": ["SDSS", "BOSS"],
        "ra": [80, 136],
        "redshift": [2, 5],
        "redshift_err": [0.0001, 0.001],
        "redshift_warning": [0, 3, 5],
        "site": ["apo"],
        "specprimary": [1],
        "spectype": ["QSO", "GALAXY"],
        "telescope": ["sloan25m"],
        "wavemax": [9210, 10500],
        "wavemin": [3800, 3900],
    }

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test all possible constraint fields with specified constraint values,
    # BOSS-DR16 only

    cons = {
        "data_release": ["BOSS-DR16"],
        "datasetgroup": ["SDSS_BOSS"],
        "dateobs_center": [
            "2018-10-07 08:44:43+00:00",
            "2018-10-07 08:44:43+00:00",
        ],
        "dec": [4, 20],
        "exptime": [2000, 12000],
        "instrument": ["SDSS", "BOSS"],
        "ra": [20, 150],
        "redshift": [0, 3],
        "redshift_err": [0.0001, 0.002],
        "redshift_warning": [0],
        "site": ["apo"],
        "specprimary": [1],
        "spectype": ["QSO", "GALAXY"],
        "telescope": ["sloan25m"],
        "wavemax": [9210, 10500],
        "wavemin": [3200, 3900],
    }

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test constraint fields with specified constraint values,
    # DESI-EDR only

    cons = {
        "data_release": ["DESI-EDR"],
        "dec": [1, 90],
        "exptime": [2000, 2500],
        "instrument": ["DESI"],
        "ra": [50, 190],
        "redshift": [1, 1.7],
        "redshift_err": [0.00006, 0.001],
        "redshift_warning": [0, 5],
        "site": ["kpno"],
        "specprimary": [1],
        "spectype": ["GALAXY", "STAR"],
        "telescope": ["kp4m"],
        "wavemax": [9700, 9900],
        "wavemin": [3500, 3700],
    }

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test constraint fields with specified constraint values,
    # all datasets

    cons = {
        "data_release": ["BOSS-DR16", "SDSS-DR16", "DESI-EDR"],
        "datasetgroup": ["SDSS_BOSS", "DESI"],
        "dec": [4, 29],
        "exptime": [2000, 12000],
        "instrument": ["SDSS", "BOSS", "DESI"],
        "ra": [80, 136],
        "redshift": [2, 5],
        "redshift_err": [0.0001, 0.001],
        "redshift_warning": [0, 3, 5],
        "site": ["apo", "kpno"],
        "specprimary": [1],
        "spectype": ["QSO", "GALAXY"],
        "telescope": ["sloan25m", "kp4m"],
    }

    found = client.find(outfields=outs, constraints=cons, limit=50)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test combo of constraint fields with specified constraint values:
    # datasetgroup, redshift, specprimary, redshift_warning, spectype

    cons = {
        "datasetgroup": ["SDSS_BOSS"],
        "redshift": [0, 0.2],
        "specprimary": [1],
        "redshift_warning": [0],
        "spectype": ["STAR"],
    }

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test combo of constraint fields with specified constraint values:
    # datasetgroup, exptime, redshift_warning

    cons = {
        "datasetgroup": ["DESI"],
        "exptime": [1000.2, 1050.9],
        "redshift_warning": [4],
    }

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test instrument and telescope constraint fields, SDSS-DR16 only

    cons = {"instrument": ["SDSS"], "telescope": ["sloan25m"]}

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test instrument and telescope constraint fields, BOSS-DR16 only

    cons = {"instrument": ["BOSS"], "telescope": ["sloan25m"]}

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test instrument and telescope constraint fields, DESI-EDR only

    cons = {"instrument": ["DESI"], "telescope": ["kp4m"]}

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test combo of constraint fields with specified constraint values:
    # site, specprimary, spectype

    cons = {"site": ["apo"], "specprimary": [1], "spectype": ["STAR"]}

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Test combo of constraint fields with specified constraint values:
    # site, specprimary, exptime

    cons = {"site": ["kpno"], "specprimary": [1], "exptime": [3000, 3010]}

    found = client.find(outfields=outs, constraints=cons, limit=10)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")


def find_outfields():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss
    # No constraints passed, no limit passed

    found = client.find(outfields=outs)
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # AUX field in only SDSS-DR16 and BOSS-DR16
    found = client.find(outfields=["run2d", "data_release"])
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # AUX field in only DESI-EDR
    found = client.find(outfields=["healpix", "data_release"])
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # AUX field in all datasets
    found = client.find(outfields=["survey", "data_release"])
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # Field does not exist in any dataset
    found = client.find(outfields=["petunia"])
    print(f"# records found: {len(found.records)}")
    print(f"First record: {found.records[0]}")
    # Field exists in at least one dataset
    found = client.find(outfields=["fiberid", "data_release"])
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # One field exists in at least one dataset, another field does not exist
    # in any datasets
    found = client.find(outfields=["plate", "dahlia", "data_release"])
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")
    # A field does not exist in the specified data_release
    try:
        found = client.find(
            outfields=["location", "data_release"],
            constraints={"data_release": ["SDSS-DR16"]},
        )
        print(f"# records found: {len(found.records)}")
        print(
            f"Datasets found: "
            "{set([f['data_release'] for f in found.records])}"
        )
        print(f"First record: {found.records[0]}")
    except Exception as e:
        print(e)
    # A field exists in only one of the specified data_releases
    found = client.find(
        outfields=["sv_primary", "data_release"],
        constraints={"data_release": ["SDSS-DR16", "DESI-EDR"]},
    )
    print(f"# records found: {len(found.records)}")
    print(f"Datasets found: {set([f['data_release'] for f in found.records])}")
    print(f"First record: {found.records[0]}")


def find_sort():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    def sortby(sort_field, cons={}, lim=500):
        found = client.find(
            outfields=outs, constraints=cons, sort=f"{sort_field}", limit=lim
        )
        print(f"Records found: {len(found.records)}")
        print(
            f"Datasets found: "
            "{set([f['data_release'] for f in found.records])}\n"
        )
        print(f"First record: {found.records[0]}\n")
        print(f"{sort_field} in first 5 records:")
        for i in range(5):
            print(found.records[i][sort_field])
        print(f"\n{sort_field} in last 5 records:")
        for i in range(495, 500):
            print(found.records[i][sort_field])

    sortby(
        "dateobs_center",
        cons={
            "dateobs_center": ["2000-05-02T23:59:59Z", "2000-08-07T03:47:16Z"]
        },
    )
    sortby("dec", cons={"dec": [4, 4.3]})
    sortby("exptime")
    sortby("ra", cons={"ra": [3, 3.3]})
    sortby("redshift", cons={"redshift": [0.9, 0.91]})
    sortby("redshift_err", cons={"redshift_err": [10.0, 20.0]})
    sortby("redshift_warning", cons={"redshift_warning": [1, 3, 5]})
    found = client.find(outfields=outs, sort="id")
    print(f"# records found: {len(found.records)}")
    print(
        f"Datasets found: {set([f['data_release'] for f in found.records])}\n"
    )
    print(f"First record: {found.records[0]}\n")
    print("sparcl_id in first 5 records:")
    for i in range(5):
        print(found.records[i].sparcl_id)
    print("\nsparcl_id in last 5 records:")
    for i in range(495, 500):
        print(found.records[i].sparcl_id)
    sortby("specid", cons={"data_release": ["DESI-EDR"]})
    sortby(
        "specprimary",
        cons={"data_release": ["BOSS-DR16", "DESI-EDR"]},
        lim=1000,
    )
    sortby("spectype")
    sortby("targetid", cons={"data_release": ["DESI-EDR"]})
    sortby("wavemax", cons={"wavemax": [10, 10000]})
    sortby(
        "wavemin",
        cons={
            "data_release": ["SDSS-DR16", "BOSS-DR16"],
            "wavemin": [1, 5000],
        },
    )
    print("\n###################### UNSORTED ######################\n")
    found = client.find(
        outfields=outs, constraints={"ra": [1, 200], "dec": [0, 80]}
    )
    print(f"Records found: {len(found.records)}")
    print(
        f"Datasets found: {set([f['data_release'] for f in found.records])}\n"
    )
    print(f"First record: {found.records[0]}\n")
    print(f"ra, dec in first 5 records:")
    for i in range(5):
        print(f"{found.records[i]['ra']}, {found.records[i]['dec']}")
    print(f"\nra, dec in last 5 records:")
    for i in range(495, 500):
        print(f"{found.records[i]['ra']}, {found.records[i]['dec']}")

    print("\n###################### SORTED ######################\n")
    found = client.find(
        outfields=outs,
        constraints={"ra": [1, 200], "dec": [0, 80]},
        sort="ra,dec",
    )
    print(f"Records found: {len(found.records)}")
    print(
        f"Datasets found: {set([f['data_release'] for f in found.records])}\n"
    )
    print(f"First record: {found.records[0]}\n")
    print(f"ra, dec in first 5 records:")
    for i in range(5):
        print(f"{found.records[i]['ra']}, {found.records[i]['dec']}")
    print(f"\nra, dec in last 5 records:")
    for i in range(495, 500):
        print(f"{found.records[i]['ra']}, {found.records[i]['dec']}")


def retrieve():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    #! client.retrieve?
    ids_lots = client.find(limit=60000).ids
    # ids from SDSS-DR16 only
    ids_sdss = client.find(
        outfields=["sparcl_id"],
        constraints={"data_release": ["SDSS-DR16"]},
        limit=3,
    ).ids

    # ids from BOSS-DR16 only
    ids_boss = client.find(
        outfields=["sparcl_id"],
        constraints={"data_release": ["BOSS-DR16"]},
        limit=3,
    ).ids

    # ids from DESI-EDR only
    ids_desi = client.find(
        outfields=["sparcl_id"],
        constraints={"data_release": ["DESI-EDR"]},
        limit=3,
    ).ids

    # ids from any dataset
    ids_any = ids_sdss + ids_boss + ids_desi


def retrieve_limit():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    #! # does the default limit work?
    client.retrieve(uuid_list=ids_lots)
    #! # does setting a limit less than the default value work?
    client.retrieve(uuid_list=ids_lots, limit=2)
    # does setting a limit over the default value but under
    #! # the maximum value work?
    client.retrieve(uuid_list=ids_lots, limit=800)
    # does trying to retrieve more than the maximum allowed
    #! # by setting limit=30000 produce the right error message?
    try:
        client.retrieve(uuid_list=ids_lots, limit=30000)
    except Exception as e:
        print(e)
    # does trying to retrieve more than the maximum allowed
    #! # by setting limit=None produce the right error message?
    try:
        client.retrieve(uuid_list=ids_lots, limit=None)
    except Exception as e:
        print(e)


def retrieve_dataset_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    # Retrieve from all datasets (dataset_list=None, the default)
    r = client.retrieve(uuid_list=ids_any, include=["data_release"])
    set([f["data_release"] for f in r.records])
    # Retrieve from only SDSS-DR16 (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any, include=["data_release"], dataset_list=["SDSS-DR16"]
    )
    set([f["data_release"] for f in r.records])
    # Retrieve from only BOSS-DR16 (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any, include=["data_release"], dataset_list=["BOSS-DR16"]
    )
    set([f["data_release"] for f in r.records])
    # Retrieve from only DESI-EDR (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any, include=["data_release"], dataset_list=["DESI-EDR"]
    )
    set([f["data_release"] for f in r.records])
    # Retrieve from only SDSS-DR16 and DESI-EDR
    # (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any,
        include=["data_release"],
        dataset_list=["SDSS-DR16", "DESI-EDR"],
    )
    set([f["data_release"] for f in r.records])


def retrieve_include():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    # include only default fields common to all datasets
    r = client.retrieve(uuid_list=ids_any, include="DEFAULT")
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields common to all datasets
    r = client.retrieve(uuid_list=ids_any, include="ALL")
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in SDSS-DR16 only
    # (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any, include="ALL", dataset_list=["SDSS-DR16"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in BOSS-DR16 only
    # (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any, include="ALL", dataset_list=["BOSS-DR16"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in DESI-EDR only
    # (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any, include="ALL", dataset_list=["DESI-EDR"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in SDSS-DR16 and DESI-EDR only
    # (should get a warning about missing ids)
    r = client.retrieve(
        uuid_list=ids_any,
        include="ALL",
        dataset_list=["SDSS-DR16", "DESI-EDR"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include only specified fields
    r = client.retrieve(
        uuid_list=ids_any, include=["sparcl_id", "flux", "redshift"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include aux fields in SDSS-DR16 only
    r = client.retrieve(
        uuid_list=ids_sdss,
        include=["run2d", "primtarget"],
        dataset_list=["SDSS-DR16"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include aux fields in BOSS-DR16 only
    r = client.retrieve(
        uuid_list=ids_boss,
        include=["fiberid", "mjd"],
        dataset_list=["BOSS-DR16"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include aux fields in BOSS-DR16 only
    r = client.retrieve(
        uuid_list=ids_any,
        include=["fiberid", "mjd"],
        dataset_list=["BOSS-DR16"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include aux fields in DESI-EDR only
    r = client.retrieve(
        uuid_list=ids_desi,
        include=["survey", "sv_primary"],
        dataset_list=["DESI-EDR"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # try to include fields that do not exist in the DB
    try:
        r = client.retrieve(
            uuid_list=ids_any, include=["sparcl_id", "magnolia"]
        )
        print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    except Exception as e:
        print(e)


def retrieve_by_specid():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    #! client.retrieve_by_specid?
    found_specids = client.find(outfields=["specid"], limit=60000)
    specids_lots = [f["specid"] for f in found_specids.records]
    # specids from SDSS-DR16 only
    found_sdss = client.find(
        outfields=["specid"],
        constraints={"data_release": ["SDSS-DR16"]},
        limit=3,
    )
    specids_sdss = [f["specid"] for f in found_sdss.records]

    # specids from BOSS-DR16 only
    found_boss = client.find(
        outfields=["specid"],
        constraints={"data_release": ["BOSS-DR16"]},
        limit=3,
    )
    specids_boss = [f["specid"] for f in found_boss.records]

    # specids from DESI-EDR only
    found_desi = client.find(
        outfields=["specid"],
        constraints={"data_release": ["DESI-EDR"]},
        limit=3,
    )
    specids_desi = [f["specid"] for f in found_desi.records]

    # specids from any dataset
    specids_any = specids_sdss + specids_boss + specids_desi


def retrieve_by_specid_limit():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    #! # does the default limit work?
    client.retrieve_by_specid(specid_list=specids_lots)
    #! # does setting a limit less than the default value work?
    client.retrieve_by_specid(specid_list=specids_lots, limit=2)
    # does setting a limit over the default value but under
    #! # the maximum value work?
    client.retrieve_by_specid(specid_list=specids_lots, limit=800)
    # does trying to retrieve more than the maximum allowed
    #! # by setting limit=30000 produce the right error message?
    try:
        client.retrieve_by_specid(specid_list=specids_lots, limit=30000)
    except Exception as e:
        print(e)
    # does trying to retrieve more than the maximum allowed
    #! # by setting limit=None produce the right error message?
    try:
        client.retrieve_by_specid(specid_list=specids_lots, limit=None)
    except Exception as e:
        print(e)


def retrieve_by_specid_dataset_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    # Retrieve from all datasets (dataset_list=None, the default)
    r = client.retrieve_by_specid(
        specid_list=specids_any, include=["data_release"]
    )
    set([f["data_release"] for f in r.records])
    # Retrieve from only SDSS-DR16
    r = client.retrieve_by_specid(
        specid_list=specids_any,
        include=["data_release"],
        dataset_list=["SDSS-DR16"],
    )
    set([f["data_release"] for f in r.records])
    # Retrieve from only BOSS-DR16
    r = client.retrieve_by_specid(
        specid_list=specids_any,
        include=["data_release"],
        dataset_list=["BOSS-DR16"],
    )
    set([f["data_release"] for f in r.records])
    # Retrieve from only DESI-EDR
    r = client.retrieve_by_specid(
        specid_list=specids_any,
        include=["data_release"],
        dataset_list=["DESI-EDR"],
    )
    set([f["data_release"] for f in r.records])
    # Retrieve from only SDSS-DR16 and DESI-EDR
    r = client.retrieve_by_specid(
        specid_list=specids_any,
        include=["data_release"],
        dataset_list=["SDSS-DR16", "DESI-EDR"],
    )
    set([f["data_release"] for f in r.records])


def retrieve_by_specid_include():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    # include only default fields common to all datasets
    r = client.retrieve_by_specid(specid_list=specids_any, include="DEFAULT")
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields common to all datasets
    r = client.retrieve_by_specid(specid_list=specids_any, include="ALL")
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in SDSS-DR16 only
    r = client.retrieve_by_specid(
        specid_list=specids_any, include="ALL", dataset_list=["SDSS-DR16"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in BOSS-DR16 only
    r = client.retrieve_by_specid(
        specid_list=specids_any, include="ALL", dataset_list=["BOSS-DR16"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in DESI-EDR only
    r = client.retrieve_by_specid(
        specid_list=specids_any, include="ALL", dataset_list=["DESI-EDR"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include all fields in SDSS-DR16 and DESI-EDR only
    r = client.retrieve_by_specid(
        specid_list=specids_any,
        include="ALL",
        dataset_list=["SDSS-DR16", "DESI-EDR"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include only specified fields
    r = client.retrieve_by_specid(
        specid_list=specids_any, include=["sparcl_id", "flux", "redshift"]
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include aux fields in SDSS-DR16 only
    r = client.retrieve_by_specid(
        specid_list=specids_sdss,
        include=["vdisp", "run2d"],
        dataset_list=["SDSS-DR16"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include aux fields in BOSS-DR16 only
    r = client.retrieve_by_specid(
        specid_list=specids_boss,
        include=["fiberid", "mjd"],
        dataset_list=["BOSS-DR16"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # include aux fields in DESI-EDR only
    r = client.retrieve_by_specid(
        specid_list=specids_desi,
        include=["survey", "sv_primary"],
        dataset_list=["DESI-EDR"],
    )
    print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    # try to include fields that do not exist in the DB
    try:
        r = client.retrieve_by_specid(
            specid_list=specids_any, include=["sparcl_id", "marigold"]
        )
        print([set(r.records[i].keys()) for i in range(len(r.records))][0])
    except Exception as e:
        print(e)


def missing():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    #! client.missing?
    some_missing_ids = ids_any + [
        "00009f84-8acc-425b-87e1-0a819f11697c",
        "00002f94-96b1-4f2d-8d1d-1a47a89fe105",
    ]


def missing_uuid_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    client.missing(uuid_list=ids_any)
    # List of 60,000 ids -- produces error
    # client.missing(uuid_list=ids_lots)
    # 2 ids not in any datasets
    client.missing(uuid_list=some_missing_ids)


def missing_dataset_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    # ids only in DESI-EDR
    client.missing(uuid_list=ids_desi, dataset_list=["SDSS-DR16", "BOSS-DR16"])
    # ids only in SDSS-DR16
    client.missing(uuid_list=ids_sdss, dataset_list=["DESI-EDR", "BOSS-DR16"])
    # ids only in BOSS-DR16
    client.missing(uuid_list=ids_boss, dataset_list=["DESI-EDR", "SDSS-DR16"])


def missing_countOnly():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    client.missing(uuid_list=some_missing_ids, countOnly=True)


def missing_specids():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    #! client.missing_specids?
    some_missing_specids = specids_any + [
        11111111111111111,
        999999999999999999,
    ]


def missing_specids_specid_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    client.missing_specids(specid_list=specids_any)
    # List of 60,000 specids -- produces error
    # client.missing_specids(specid_list=specids_lots)
    # 2 specids not in any datasets
    client.missing_specids(specid_list=some_missing_specids)


def missing_specids_dataset_list():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    # specids only in DESI-EDR
    client.missing_specids(
        specid_list=specids_desi, dataset_list=["SDSS-DR16", "BOSS-DR16"]
    )
    # specids only in SDSS-DR16
    client.missing_specids(
        specid_list=specids_sdss, dataset_list=["DESI-EDR", "BOSS-DR16"]
    )
    # specids only in BOSS-DR16
    client.missing_specids(
        specid_list=specids_boss, dataset_list=["DESI-EDR", "SDSS-DR16"]
    )


def missing_specids_countOnly():
    global client, ids_any, ids_boss, ids_desi, ids_lots, ids_sdss, outs
    global some_missing_ids, some_missing_specids, specids_any, specids_boss
    global specids_desi, specids_lots, specids_sdss

    client.missing_specids(specid_list=some_missing_specids, countOnly=True)


###
###########################################################################

###########################################################################
# Tests
#


# START Pre-notebook-sections code
class NotebookSectionTest(unittest.TestCase):
    """Test each section in notebook: SPARCL_client_method_tests.ipynb"""

    def test000_get_all_fields(self):
        actual = get_all_fields()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test001_get_all_fields_dataset_list(self):
        actual = get_all_fields_dataset_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test002_get_default_fields(self):
        actual = get_default_fields()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test003_get_default_fields_dataset_list(self):
        actual = get_default_fields_dataset_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test004_get_available_fields(self):
        actual = get_available_fields()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test005_get_available_fields_dataset_list(self):
        actual = get_available_fields_dataset_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test006_find(self):
        actual = find()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test007_find_limit(self):
        actual = find_limit()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test008_find_constraints(self):
        actual = find_constraints()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test009_find_outfields(self):
        actual = find_outfields()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test010_find_sort(self):
        actual = find_sort()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test011_retrieve(self):
        actual = retrieve()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test012_retrieve_limit(self):
        actual = retrieve_limit()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test013_retrieve_dataset_list(self):
        actual = retrieve_dataset_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test014_retrieve_include(self):
        actual = retrieve_include()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test015_retrieve_by_specid(self):
        actual = retrieve_by_specid()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test016_retrieve_by_specid_limit(self):
        actual = retrieve_by_specid_limit()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test017_retrieve_by_specid_dataset_list(self):
        actual = retrieve_by_specid_dataset_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test018_retrieve_by_specid_include(self):
        actual = retrieve_by_specid_include()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test019_missing(self):
        actual = missing()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test020_missing_uuid_list(self):
        actual = missing_uuid_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test021_missing_dataset_list(self):
        actual = missing_dataset_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test022_missing_countOnly(self):
        actual = missing_countOnly()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test023_missing_specids(self):
        actual = missing_specids()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test024_missing_specids_specid_list(self):
        actual = missing_specids_specid_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test025_missing_specids_dataset_list(self):
        actual = missing_specids_dataset_list()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")

    def test026_missing_specids_countOnly(self):
        actual = missing_specids_countOnly()
        expected = None
        self.assertEqual(actual, expected, msg="Actual to Expected")



# Python library
#!import logging
#!import os
import datetime
import time
import socket
from inspect import cleandoc

# External packages
#   none
# LOCAL packages
#   none


# e.g. pdocstr(coexist_radec.__doc__),
def pdocstr(docstring):
    return cleandoc(docstring).replace("\n", " ")


def tic():
    tic.start = time.perf_counter()


def toc():
    elapsed_seconds = time.perf_counter() - tic.start
    return elapsed_seconds  # fractional


def here_now():
    hostname = socket.gethostname()
    now = str(datetime.datetime.now())
    return (hostname, now)


# dict((k,len(v)) for (k,v) in qs[0]['spzline'].items())
def objform(obj):
    """Nested structure of python object. Avoid spewing big lists."""
    if obj is None:
        return None
    elif type(obj) is list:
        return f"CNT={len(obj)}"
    elif type(obj) is dict:
        return dict((k, objform(v)) for (k, v) in obj.items())
    else:
        return str(type(obj))



import traceback


def genSparclException(response, verbose=False):
    """Given status from Server response.json(), which is a dict, generate
    a native SPARCL exception suitable for Science programs."""

    content = response.content
    if verbose:
        print(f"Exception: response content={content}")
    status = response.json()

    # As of Python 3.10.0.alpha6, python "match" statement could be used
    # instead of if-elif-else.
    # https://docs.python.org/3.10/whatsnew/3.10.html#pep-634-structural-pattern-matching

    if status.get("errorCode") == "BADPATH":
        return BadPath(status.get("errorMessage"))
    elif response.status_code == 429:  # too many requests, throttled
        return TooManyRequests(status.get("detail"))
    elif status.get("errorCode") == "BADQUERY":
        return BadQuery(status.get("errorMessage"))
    elif status.get("errorCode") == "UNKFIELD":
        return UnknownField(status.get("errorMessage"))
    elif status.get("errorCode") == "BADCONST":
        return BadSearchConstraint(status.get("errorMessage"))
    elif status.get("errorCode") == "NODRACCESS":
        return AccessNotAllowed(status.get("errorMessage"))
    else:
        return UnknownServerError(
            f"{status.get('errorMessage')} "
            f"[{status.get('errorCode')}] "
            f"{status.get('traceback')}"
        )


class BaseSparclException(Exception):
    """Base Class for all SPARCL exceptions."""

    error_code = "UNKNOWN"
    error_message = "<NA>"
    traceback = True

    def get_subclass_name(self):
        return self.__class__.__name__

    def __init__(self, error_message, error_code=None):
        Exception.__init__(self)
        self.error_message = error_message
        if error_code:
            self.error_code = error_code
        self.traceback = traceback.format_exc()

    def __str__(self):
        return f"[{self.error_code}] {self.error_message}"

    def to_dict(self):
        """Convert a SPARCL exception to a python dictionary"""
        dd = dict(errorMessage=self.error_message, errorCode=self.error_code)
        if self.traceback is not None:
            dd["traceback"] = self.traceback
        return dd


class BadPath(BaseSparclException):
    """A field path starts with a non-core field."""

    error_code = "BADPATH"

class TooManyRequests(BaseSparclException):

    error_code = "TOOMANYREQUESTS"

class BadQuery(BaseSparclException):
    """Bad find constraints."""

    error_code = "BADPATH"


class BadInclude(BaseSparclException):
    """Include list contains invalid data field(s)."""

    error_code = "BADINCL"


class UnknownServerError(BaseSparclException):
    """Client got a status response from the SPARC Server that we do not
    know how to decode."""

    error_code = "UNKNOWN"


class UnkDr(BaseSparclException):
    """The Data Release is not known or not supported."""

    error_code = "UNKDR"


class ReadTimeout(BaseSparclException):
    """The server did not send any data in the allotted amount of time."""

    error_code = "RTIMEOUT"


class UnknownSparcl(BaseSparclException):
    """Unknown SPARCL error.  If this is ever raised (seen in a log)
    create and use a new BaseSparcException exception that is more specific."""

    error_code = "UNKSPARC"


class UnknownField(BaseSparclException):
    """Unknown field name for a record"""

    error_code = "UNKFIELD"


class NoCommonIdField(BaseSparclException):
    """The field name for Science id field is not common to all Data Sets"""

    error_code = "IDNOTCOM"


class ServerConnectionError(BaseSparclException):
    error_code = "SRVCONER"


class BadSearchConstraint(BaseSparclException):
    error_code = "BADSCONS"


class NoRecords(BaseSparclException):
    """Results did not contain any records"""

    error_code = "NORECORD"


class TooManyRecords(BaseSparclException):
    """Too many records asked for in RETRIEVE"""

    error_code = "TOOMANYR"


class NoIDs(BaseSparclException):
    """The length of the list of original IDs passed to the reorder
    method was zero"""

    error_code = "NOIDS"


class AccessNotAllowed(BaseSparclException):
    error_code = "DSDENIED"


# error_code values should be no bigger than 8 characters 12345678



# Python Standard Library
from abc import ABC, abstractmethod
import copy

#!from pprint import pformat
from enum import Enum, auto

# External Packages
import numpy as np

#!import pandas as pd
from specutils import Spectrum1D
import astropy.units as u
from astropy.nddata import InverseVariance

# Local Packages
import sparcl.exceptions as ex


"""It would be much better if this were abstracted and easier to
update with new Data Types and new DataReleases.  Perhaps use
something like the Server "Personalities".  I've rejected abstracting
for now because I think we need operational experience with the DataType
feature and how it interacts with other features (especially global
Rename and retrieve(INCLUDE).

DataType conversion should be done completely within the Client, not
on the Server.  The obvious reason is Clients are language
dependent, the Server API is not.  But for Client to be able to know
all about fields names (mapping from original to new names, which ones
are required) it needs info from the Server.  The Client gets such
tables on instance instantiation through one web-service call that
grabs everything that pulls it appart into multiple DataField related
LUTs (LookUpTables, aka dictionaries).

Questions abound for use-cases.

1. Is it very important to be able to convert a record LIST to a
   single data structure? Example: for Pandas DataFrame we combine all
   vectors in a record into a 2D DataFrame.  What about those across
   records into a 3D DataFrame?

2. Should vectors and scalars be funadmentally separated? (see #1) If
   so, how do we avoid hard coding the distinction for every
   DataRelease?

"""


# Replace all uses of string rtype with enum @@@
class Rtype(Enum):
    JSON = auto()
    NUMPY = auto()
    PANDAS = auto()
    SPECTRUM1D = auto()


class Convert(ABC):
    """Convert JSON record to mix of plain python
    and selected data record type.
    """

    @abstractmethod
    def to_numpy(self, record, o2nLUT):
        newrec = copy.deepcopy(record)
        return newrec

    @abstractmethod
    def to_spectrum1d(self, record, o2nLUT):
        newrec = copy.deepcopy(record)
        return newrec


#!    @abstractmethod
#!    def to_pandas(self, record, o2nLUT):
#!        newrec = copy.deepcopy(record)
#!        return(newrec)


class NoopConvert(Convert):
    def to_numpy(self, record, o2nLUT):
        return record

    def to_spectrum1d(self, record, o2nLUT):
        return record

    def to_pandas(self, record, o2nLUT):
        return record


class SdssDr16(Convert):
    def to_numpy(self, record, o2nLUT):
        """Convert FitsFile record to a structure that uses Numpy"""
        arflds = [
            "spectra.coadd.and_mask",
            "spectra.coadd.flux",
            "spectra.coadd.ivar",
            "spectra.coadd.loglam",
            "spectra.coadd.model",
            "spectra.coadd.or_mask",
            "spectra.coadd.sky",
            "spectra.coadd.wdisp",
        ]
        lofl = [record[o2nLUT[f]] for f in arflds if f in o2nLUT]
        newrec = dict(nparr=np.array(lofl))
        for orig, new in o2nLUT.items():
            if orig in arflds:
                continue
            if new in record:
                newrec[new] = record[new]
        return newrec

    # Sdss
    def to_spectrum1d(self, record, o2nLUT):
        arflds = [
            "red_shift",
            "spectra.coadd.flux",
            "spectra.coadd.ivar",
            "spectra.coadd.loglam",
            "spectra.coadd.and_mask",
        ]

        loglam = record[o2nLUT["spectra.coadd.loglam"]]
        flux = record[o2nLUT["spectra.coadd.flux"]]
        ivar = record[o2nLUT["spectra.coadd.ivar"]]
        and_mask = record[o2nLUT["spectra.coadd.and_mask"]]

        wavelength = (10 ** np.array(loglam)) * u.AA
        flux = np.array(flux) * 10**-17 * u.Unit("erg cm-2 s-1 AA-1")
        ivar = InverseVariance(np.array(ivar))
        z = record.get("red_shift")

        newrec = dict(
            # flux, uncertainty, wavevelength, mask(and), redshift
            spec1d=Spectrum1D(
                spectral_axis=wavelength,
                flux=flux,
                uncertainty=ivar,
                redshift=z,
                mask=and_mask,
            ),
        )
        for orig, new in o2nLUT.items():
            if orig in arflds:
                continue
            if new in record:
                newrec[new] = record[new]

        return newrec


#!    def to_pandas(self, record, o2nLUT):
#!        arflds = [
#!            'spectra.coadd.and_mask',
#!            'spectra.coadd.flux',
#!            'spectra.coadd.ivar',
#!            'spectra.coadd.loglam',
#!            'spectra.coadd.model',
#!            'spectra.coadd.or_mask',
#!            'spectra.coadd.sky',
#!            'spectra.coadd.wdisp',
#!            ]
#!        dfdict = dict((o2nLUT[f], record[o2nLUT[f]])
#!                      for f in arflds if f in o2nLUT)
#!        newrec = dict(df = pd.DataFrame(dfdict))
#!        for orig,new in o2nLUT.items():
#!            if orig in arflds:
#!                continue
#!            if new in record:
#!                newrec[new] = record[new]
#!        return(newrec)


class BossDr16(Convert):
    def to_numpy(self, record, o2nLUT):
        arflds = [
            "spectra.coadd.AND_MASK",
            "spectra.coadd.FLUX",
            "spectra.coadd.IVAR",
            "spectra.coadd.LOGLAM",
            "spectra.coadd.MODEL",
            "spectra.coadd.OR_MASK",
            "spectra.coadd.SKY",
            "spectra.coadd.WDISP",
        ]
        lofl = [record[o2nLUT[f]] for f in arflds if f in o2nLUT]
        newrec = dict(nparr=np.array(lofl))
        for orig, new in o2nLUT.items():
            # Don't carry over the fields used to build the new datatype.
            # This would be duplication since their content is already
            # in the new datatype.
            if orig in arflds:
                continue
            if new in record:
                newrec[new] = record[new]
        return newrec

    # BOSS
    def to_spectrum1d(self, record, o2nLUT):
        arflds = [
            "red_shift",
            "spectra.coadd.FLUX",
            "spectra.coadd.IVAR",
            "spectra.coadd.LOGLAM",
            "spectra.coadd.AND_MASK",
        ]
        loglam = record[o2nLUT["spectra.coadd.LOGLAM"]]
        flux = record[o2nLUT["spectra.coadd.FLUX"]]
        ivar = record[o2nLUT["spectra.coadd.IVAR"]]
        and_mask = record[o2nLUT["spectra.coadd.AND_MASK"]]

        wavelength = (10 ** np.array(loglam)) * u.AA
        flux = np.array(flux) * 10**-17 * u.Unit("erg cm-2 s-1 AA-1")
        ivar = InverseVariance(np.array(ivar))
        z = record.get("red_shift")

        newrec = dict(
            # flux, uncertainty, wavelength, mask(and), redshift
            spec1d=Spectrum1D(
                spectral_axis=wavelength,
                flux=flux,
                uncertainty=ivar,
                redshift=z,
                mask=and_mask,
            ),
        )
        for orig, new in o2nLUT.items():
            if orig in arflds:
                continue
            if new in record:
                newrec[new] = record[new]
        return newrec


#!    def to_pandas(self, record, o2nLUT): # BOSS
#!        arflds = [
#!            'spectra.coadd.AND_MASK',
#!            'spectra.coadd.FLUX',
#!            'spectra.coadd.IVAR',
#!            'spectra.coadd.LOGLAM',
#!            'spectra.coadd.MODEL',
#!            'spectra.coadd.OR_MASK',
#!            'spectra.coadd.SKY',
#!            'spectra.coadd.WDISP',
#!            ]
#!        dfdict = dict((o2nLUT[f], record[o2nLUT[f]])
#!                      for f in arflds if f in o2nLUT)
#!        newrec = dict(df = pd.DataFrame(dfdict))
#!        for orig,new in o2nLUT.items():
#!            if orig in arflds:
#!                continue
#!            if new in record:
#!                newrec[new] = record[new]
#!        return(newrec)


class Desi(Convert):
    def to_numpy(self, record, o2nLUT):
        arflds = [
            "spectra.b_flux",
            "spectra.b_ivar",
            "spectra.b_mask",
            "spectra.b_wavelength",
            "spectra.r_flux",
            "spectra.r_ivar",
            "spectra.r_mask",
            "spectra.r_wavelength",
            "spectra.z_flux",
            "spectra.z_ivar",
            "spectra.z_mask",
            "spectra.z_wavelength",
        ]
        lofl = [record[o2nLUT[f]] for f in arflds if f in o2nLUT]
        newrec = dict(nparr=np.array(lofl))
        for orig, new in o2nLUT.items():
            if orig in arflds:
                continue
            if new in record:
                newrec[new] = record[new]
        return newrec

    def to_spectrum1d(self, record, o2nLUT):  # Desi
        arflds = [
            "red_shift",
            "spectra.b_flux",
            "spectra.b_ivar",
            "spectra.b_mask",
            "spectra.b_wavelength",
            "spectra.r_flux",
            "spectra.r_ivar",
            "spectra.r_mask",
            "spectra.r_wavelength",
            "spectra.z_flux",
            "spectra.z_ivar",
            "spectra.z_mask",
            "spectra.z_wavelength",
        ]

        z = record.get("red_shift")

        # _b
        wavelength_b = record[o2nLUT["spectra.b_wavelength"]]
        flux_b = record[o2nLUT["spectra.b_flux"]]
        ivar_b = record[o2nLUT["spectra.b_ivar"]]
        mask_b = record[o2nLUT["spectra.b_mask"]]
        # Define units
        wavelength_b = np.array(wavelength_b) * u.AA
        flux_b = np.array(flux_b) * 10**-17 * u.Unit("erg cm-2 s-1 AA-1")
        ivar_b = InverseVariance(np.array(ivar_b))

        # _r
        wavelength_r = record[o2nLUT["spectra.r_wavelength"]]
        flux_r = record[o2nLUT["spectra.r_flux"]]
        ivar_r = record[o2nLUT["spectra.r_ivar"]]
        mask_r = record[o2nLUT["spectra.r_mask"]]
        # Define units
        wavelength_r = np.array(wavelength_r) * u.AA
        flux_r = np.array(flux_r) * 10**-17 * u.Unit("erg cm-2 s-1 AA-1")
        ivar_r = InverseVariance(np.array(ivar_r))

        # _z
        wavelength_z = record[o2nLUT["spectra.z_wavelength"]]
        flux_z = record[o2nLUT["spectra.z_flux"]]
        ivar_z = record[o2nLUT["spectra.z_ivar"]]
        mask_z = record[o2nLUT["spectra.z_mask"]]
        # Define units
        wavelength_z = np.array(wavelength_z) * u.AA
        flux_z = np.array(flux_z) * 10**-17 * u.Unit("erg cm-2 s-1 AA-1")
        ivar_z = InverseVariance(np.array(ivar_z))

        newrec = dict(
            # flux, uncertainty, wavevelength, mask, redshift
            b_spec1d=Spectrum1D(
                spectral_axis=wavelength_b,
                flux=flux_b,
                uncertainty=ivar_b,
                redshift=z,
                mask=mask_b,
            ),
            r_spec1d=Spectrum1D(
                spectral_axis=wavelength_r,
                flux=flux_r,
                uncertainty=ivar_r,
                redshift=z,
                mask=mask_r,
            ),
            z_spec1d=Spectrum1D(
                spectral_axis=wavelength_z,
                flux=flux_z,
                uncertainty=ivar_z,
                redshift=z,
                mask=mask_z,
            ),
        )
        for orig, new in o2nLUT.items():
            if orig in arflds:
                continue
            if new in record:
                newrec[new] = record[new]
        return newrec


class DesiDenali(Desi):
    pass


class DesiEverest(Desi):
    pass


# DR Instance LookUp Table
diLUT = {
    "SDSS-DR16": SdssDr16(),
    "BOSS-DR16": BossDr16(),
    "DESI-denali": DesiDenali(),
    "DESI-everest": DesiEverest(),
    #'Unknown': NoopConvert(),
}


def convert(record, rtype, client, include, verbose=False):
    if rtype is None:
        return record

    dr = record["_dr"]

    # Validate parameters
    if dr not in diLUT:
        allowed = ", ".join(list(diLUT.keys()))
        msg = (
            f'The Data Set associated with a records, "{dr}",'
            f" is not supported for Type Conversion."
            f" Available Data Sets are: {allowed}."
        )
        raise ex.UnkDr(msg)

    drin = diLUT.get(dr, NoopConvert())

    o2nLUT = copy.copy(client.orig2newLUT[dr])  # orig2newLUT[dr][orig] = new
    o2nLUT["_dr"] = "_dr"
    #!n2oLUT = client.new2origLUT[dr]
    #!required = set(client.required[dr])
    #!if include is not None:
    #!    nuke = set(n2oLUT.keys()).difference(required.union(include))
    #!    for new in nuke:
    #!        del o2nLUT[n2oLUT[new]]

    if rtype == "json":
        return record
    elif rtype == "numpy":
        return drin.to_numpy(record, o2nLUT)
    elif rtype == "pandas":
        return drin.to_pandas(record, o2nLUT)
    elif rtype == "spectrum1d":
        return drin.to_spectrum1d(record, o2nLUT)
    else:
        raise Exception(f"Unknown record type ({rtype})")
    return None



"""Client module for SPARCL.
This module interfaces to the SPARC-Server to get spectra data.
"""

# python -m unittest tests.tests_api

# ### Run tests against DEV
# serverurl=http://localhost:8050 python -m unittest tests.tests_api
#
#  ### Run tests Against PAT Server.
#  export serverurl=https://sparc1.datalab.noirlab.edu/
#  python -m unittest tests.tests_api

#
# Doctest example:
#   cd ~/sandbox/sparclclient
#   activate
#   python sparcl/client.py
#   ## Returns NOTHING if everything works, else lists errors.

############################################
# Python Standard Library
from urllib.parse import urlencode, urlparse
from warnings import warn
import pickle
import getpass
import datetime
import re

#!from pathlib import Path
import tempfile

############################################
# External Packages
import requests
import jwt

#!from requests.auth import HTTPBasicAuth
from requests.auth import AuthBase

############################################
# Local Packages
from sparcl.fields import Fields
import sparcl.utils as ut
import sparcl.exceptions as ex

#!import sparcl.type_conversion as tc
from sparcl import __version__
from sparcl.Results import Found, Retrieved


MAX_CONNECT_TIMEOUT = 3.1  # seconds
MAX_READ_TIMEOUT = 150 * 60  # seconds
MAX_NUM_RECORDS_RETRIEVED = int(24e3)  # Minimum Hard Limit = 25,000
#!MAX_NUM_RECORDS_RETRIEVED = int(5e4) #@@@ Reduce !!!


_pat_hosts = [
    "sparc1.datalab.noirlab.edu",
    "sparc2.datalab.noirlab.edu",
    "astrosparcl.datalab.noirlab.edu",
]

# Upload to PyPi:
#   python3 -m build --wheel
#   twine upload dist/*

# Use Google Style Python Docstrings so autogen of Sphinx doc works:
#  https://www.sphinx-doc.org/en/master/usage/extensions/example_google.html
#  https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html
#
# Use sphinx-doc emacs minor mode to insert docstring skeleton.
# C-c M-d in function/method def

# ### Generate documentation:
# cd ~/sandbox/sparclclient
# sphinx-apidoc -f -o source sparcl
# make html
# firefox -new-tab "`pwd`/build/html/index.html"

# Using HTTPie (http://httpie.org):
# http :8030/sparc/version

# specids = [394069118933821440, 1355741587413428224, 1355617892355303424,
#   1355615143576233984, 1355661872820414464, 1355755331308775424,
#   1355716848401803264]
# client = client.SparclClient(url='http://localhost:8030/sparc')
# client.retrieve(specids)[0].keys() # >> dict_keys(['flux','loglam'])
#
# data0 = client.retrieve(specids,columns='flux')
# f'{len(str(data0)):,}'   # -> '3,435,687'
#
# dataall = client.retrieve(specids,columns=allc)
# f'{len(str(dataall)):,}' # -> '27,470,052'

_PROD = "https://astrosparcl.datalab.noirlab.edu"  # noqa: E221
_STAGE = "https://sparclstage.datalab.noirlab.edu"  # noqa: E221
_PAT = "https://sparc1.datalab.noirlab.edu"  # noqa: E221
_DEV = "http://localhost:8050"  # noqa: E221


# client_version = pkg_resources.require("sparclclient")[0].version
client_version = __version__

DEFAULT = "DEFAULT"
ALL = "ALL"
RESERVED = set([DEFAULT, ALL])


###########################
# ## Convenience Functions

# Following can be done with:
#   set.intersection(*sets)
#
#!def intersection(*lists):
#!    """Return intersection of all LISTS."""
#!    return set(lists[0]).intersection(*lists[1:])


class TokenAuth(AuthBase):
    """Attaches HTTP Token Authentication to the given Request object."""

    def __init__(self, token, renew_check):
        # setup any auth-related data here
        self.token = token
        self.renew_check = renew_check  # SparcClient Method

    def __call__(self, request):
        # modify and return the request
        # check and renew token if needed
        self.renew_check(renew=True)
        request.headers["Authorization"] = self.token
        return request


###########################
# ## The Client class


class SparclClient:  # was SparclApi()
    """Provides interface to SPARCL Server.
    When using this to report a bug, set verbose to True. Also print
    your instance of this.  The results will include important info
    about the Client and Server that is usefule to Developers.

    Args:
        url (:obj:`str`, optional): Base URL of SPARCL Server. Defaults
            to 'https://astrosparcl.datalab.noirlab.edu'.

        verbose (:obj:`bool`, optional): Default verbosity is set to
            False for all client methods.

        connect_timeout (:obj:`float`, optional): Number of seconds to
            wait to establish connection with server. Defaults to
            1.1.

        read_timeout (:obj:`float`, optional): Number of seconds to
            wait for server to send a response. Generally time to
            wait for first byte. Defaults to 5400.

        announcement (:obj:`bool`, optional): SPARCL announcements.
            Defaults to True.

    Example:
        >>> client = SparclClient()

    Raises:
        Exception: Object creation compares the version from the
            Server against the one expected by the Client. Throws an
            error if the Client is a major version or more behind.

    """

    KNOWN_GOOD_API_VERSION = 12.0  # @@@ Change when Server version incremented

    def __init__(
        self,
        *,
        url=_PROD,
        verbose=False,
        show_curl=False,
        connect_timeout=1.1,  # seconds
        read_timeout=90 * 60,  # seconds
        announcement=True,
    ):
        """Create client instance."""
        session = requests.Session()
        self.session = session
        self.session.auth = None
        self.rooturl = url.rstrip("/")  # eg. "http://localhost:8050"
        self.apiurl = f"{self.rooturl}/sparc"
        self.apiversion = None
        self.token = None
        self.refresh_token = None
        self.token_exp = None
        self.verbose = verbose
        self.show_curl = show_curl  # Show CURL equivalent of client method
        #!self.internal_names = internal_names
        self.c_timeout = min(
            MAX_CONNECT_TIMEOUT, float(connect_timeout)
        )  # seconds
        self.r_timeout = min(MAX_READ_TIMEOUT, float(read_timeout))  # seconds

        # require response within this num seconds
        # https://2.python-requests.org/en/master/user/advanced/#timeouts
        # (connect timeout, read timeout) in seconds
        self.timeout = (self.c_timeout, self.r_timeout)
        # @@@ read timeout should be a function of the POST payload size
        self.announcement = announcement

        if verbose:
            print(f"apiurl={self.apiurl}")

        if announcement:
            ann = f"{self.apiurl}/announcement/"
            annres = requests.get(ann, timeout=self.timeout).content
            print(f"announcement={annres.json()['announcement']}")
            #print(f"announcement={annstr.decode('utf-8')['announcement']}")

        # Get API Version
        try:
            endpoint = f"{self.apiurl}/version/"
            verstr = requests.get(endpoint, timeout=self.timeout).content
        except requests.ConnectionError as err:
            msg = f"Could not connect to {endpoint}. {str(err)}"
            if urlparse(url).hostname in _pat_hosts:
                msg += "Did you enable VPN?"
            raise ex.ServerConnectionError(msg) from None  # disable chaining

        self.apiversion = float(verstr)

        expected_api = SparclClient.KNOWN_GOOD_API_VERSION
        if (int(self.apiversion) - int(expected_api)) >= 1:
            msg = (
                f"The SPARCL Client you are running expects an older "
                f"version of the API services. "
                f'Please upgrade to the latest "sparclclient" using  '
                f'"pip install --upgrade sparclclient". '
                f"The Client you are using expected version "
                f"{SparclClient.KNOWN_GOOD_API_VERSION} but got "
                f"{self.apiversion} from the SPARCL Server "
                f"at {self.apiurl}."
            )
            raise Exception(msg)

        self.clientversion = client_version
        self.fields = Fields(self.apiurl)

        ###
        ####################################################
        # END __init__()

    def __repr__(self):
        #!f' internal_names={self.internal_names},'
        return (
            f"(sparclclient:{self.clientversion},"
            f" api:{self.apiversion},"
            f" {self.apiurl},"
            f" client_hash={ut.githash()},"
            f" verbose={self.verbose},"
            f" connect_timeout={self.c_timeout},"
            f" read_timeout={self.r_timeout}),"
            f" announcement={self.announcement})"
        )

    def token_expired(self, renew=False):
        """
            POST http://localhost:8050/sparc/renew_token/
            Content-Type: application/json
            {
              "refresh_token": "..."
            }

            Returns an 'access' token
        """
        now = datetime.datetime.now()
        expired = True
        if self.token_exp is None or self.token_exp <= now:
            expired = True
        else:
            expired = False

        if expired and renew:
            url = f"{self.apiurl}/renew_token/"
            resp = requests.post(url, json={"refresh_token": self.renew_token})
            resp.raise_for_status()
            data = resp.json()
            #@print(f"{data=}")
            self.token = data['access']
            self.set_token_exp()
            expired = False

        return expired

    def login(self, email, password=None):
        """Login to the SPARCL service.

        Args:
            email (:obj:`str`): User login email.

            password (:obj:`str`, optional): User SSO password.
                If not given, the output will prompt the user
                to enter in their SSO password.

        Returns:
            None.

        Example:
            >>>
            >> client = SparclClient()
            >> client.login('test_user@noirlab.edu', 'testpw')
            Logged in successfully with email='test_user@noirlab.edu'
        """

        if email is None:  # "logout"
            old_email = self.session.auth[0] if self.session.auth else None
            self.session.auth = None
            self.token = None
            print(
                f"Logged-out successfully. "
                f" Previously logged-in with email {old_email}."
            )
            return None
        if password is None:
            password = getpass.getpass(prompt="SSO Password: ")
        url = f"{self.apiurl}/get_token/"
        # print(f'login: get_token {url=}')
        res = requests.post(
            url,
            json=dict(email=email, password=password),
            timeout=self.timeout,
        )
        try:
            res.raise_for_status()
            #!print(f"DBG: {res.content=}")
            self.token = res.json()['access']
            self.renew_token = res.json()['refresh']
            self.session.auth = (email, password)
        except Exception:
            self.session.auth = None
            self.token = None
            self.renew_token = None
            self.token_exp = None
            msg = (
                "Could not login with given credentials."
                ' Reverted to "Anonymous" user.'
            )
            return msg

        self.set_token_exp()
        print(f"Logged in successfully with email={email}")
        return None

    def set_token_exp(self):
        decoded = jwt.decode(
            self.token,
            algorithms=['HS256',],
            options={'verify_signature': False}
        )
        self.token_exp = datetime.datetime.fromtimestamp(decoded['exp'])

    def logout(self):
        """Logout of the SPARCL service.

        Args:
            None.

        Returns:
            None.

        Example:
            >>> client = SparclClient()
            >>> client.logout()
            Logged-out successfully.  Previously logged-in with email None.
        """

        return self.login(None)

    @property
    def authorized(self):
        auth = TokenAuth(self.token, self.token_expired) if self.token else None  # noqa: E501
        response = requests.get(
            f"{self.apiurl}/auth_status/", auth=auth, timeout=self.timeout
        )
        auth_status = response.json()
        #! print(f"DBG authorized: {auth_status=}")

        username = auth_status.get("Loggedin_User")
        #! all_private_drs = set(auth_status.get("All_Private_Datasets"))
        all_public_drs = set(auth_status.get("All_Public_Datasets"))
        auth_drs = set(auth_status.get("Authorized_Private_Datasets"))
        res = dict(
            Loggedin_As=username,  # email
            Authorized_Datasets=auth_drs | all_public_drs,
            #! Unauthorized_Datasets=all_private_drs - auth_drs,
            #! All_Private_Datasets=all_private_drs,
            #! All_Datasets=all_drs,
        )
        return res

    @property
    def all_datasets(self):
        """Set of all DataSets available from Server"""
        return self.fields.all_drs

    def get_default_fields(self, *, dataset_list=None):
        """Get fields tagged as 'default' that are in DATASET_LIST.
        These are the fields used for the DEFAULT value of the include
        parameter of client.retrieve().

        Args:
            dataset_list (:obj:`list`, optional): List of data sets from
                which to get the default fields. Defaults to None, which
                will return the intersection of default fields in all
                data sets hosted on the SPARCL database.

        Returns:
            List of fields tagged as 'default' from DATASET_LIST.

        Example:
            >>> client = SparclClient()
            >>> client.get_default_fields()
            ['dec', 'flux', 'ra', 'sparcl_id', 'specid', 'wavelength']
        """

        if dataset_list is None:
            dataset_list = self.fields.all_drs

        assert isinstance(
            dataset_list, (list, set)
        ), f"DATASET_LIST must be a list. Found {dataset_list}"

        common = set(self.fields.common(dataset_list))
        union = self.fields.default_retrieve_fields(dataset_list=dataset_list)
        return sorted(common.intersection(union))

    def get_all_fields(self, *, dataset_list=None):
        """Get fields tagged as 'all' that are in DATASET_LIST.
        These are the fields used for the ALL value of the include parameter
        of client.retrieve().

        Args:
            dataset_list (:obj:`list`, optional): List of data sets from
                which to get all fields. Defaults to None, which
                will return the intersection of all fields in all
                data sets hosted on the SPARCL database.

        Returns:
            List of fields tagged as 'all' from DATASET_LIST.

        Example:
            >>> client = SparclClient()
            >>> client.get_all_fields()
            ['data_release', 'datasetgroup', 'dateobs', 'dateobs_center', 'dec', 'exptime', 'flux', 'instrument', 'ivar', 'mask', 'model', 'ra', 'redshift', 'redshift_err', 'redshift_warning', 'site', 'sparcl_id', 'specid', 'specprimary', 'spectype', 'survey', 'targetid', 'telescope', 'wave_sigma', 'wavelength', 'wavemax', 'wavemin']
        """  # noqa: E501

        common = set(self.fields.common(dataset_list))
        union = self.fields.all_retrieve_fields(dataset_list=dataset_list)
        return sorted(common.intersection(union))

    def _validate_science_fields(self, science_fields, *, dataset_list=None):
        """Raise exception if any field name in SCIENCE_FIELDS is
        not registered in at least one of DATASET_LIST."""
        if dataset_list is None:
            dataset_list = self.fields.all_drs
        all = set(self.fields.common(dataset_list=dataset_list))
        unk = set(science_fields) - all
        if len(unk) > 0:
            drs = self.fields.all_drs if dataset_list is None else dataset_list
            msg = (
                f'Unknown fields "{",".join(unk)}" given '
                f'for DataSets {",".join(drs)}. '
                f'Allowed fields are: {",".join(all)}. '
            )
            raise ex.UnknownField(msg)
        return True

    def _common_internal(self, *, science_fields=None, dataset_list=None):
        self._validate_science_fields(
            science_fields, dataset_list=dataset_list
        )

        if dataset_list is None:
            dataset_list = self.fields.all_drs
        if science_fields is None:
            science_fields = self.fields.all_fields
        common = self.fields.common_internal(dataset_list)
        flds = set()
        for dr in dataset_list:
            for sn in science_fields:
                flds.add(self.fields._internal_name(sn, dr))
        return common.intersection(flds)

    # Return Science Field Names (not Internal)
    def get_available_fields(self, *, dataset_list=None):
        """Get subset of fields that are in all (or selected) DATASET_LIST.
        This may be a bigger list than will be used with the ALL keyword to
        client.retreive().

        Args:
            dataset_list (:obj:`list`, optional): List of data sets from
                which to get available fields. Defaults to None, which
                will return the intersection of all available fields in
                all data sets hosted on the SPARCL database.

        Returns:
            Set of fields available from data sets in DATASET_LIST.

        Example:
            >>> client = SparclClient()
            >>> sorted(client.get_available_fields())
            ['data_release', 'datasetgroup', 'dateobs', 'dateobs_center', 'dec', 'exptime', 'extra_files', 'file', 'flux', 'instrument', 'ivar', 'mask', 'model', 'ra', 'redshift', 'redshift_err', 'redshift_warning', 'site', 'sparcl_id', 'specid', 'specprimary', 'spectype', 'survey', 'targetid', 'telescope', 'updated', 'wave_sigma', 'wavelength', 'wavemax', 'wavemin']

        """  # noqa: E501

        drs = self.fields.all_drs if dataset_list is None else dataset_list
        every = [set(self.fields.n2o[dr]) for dr in drs]
        return set.intersection(*every)

    @property
    def version(self):
        """Return version of Server Rest API used by this client.
        If the Rest API changes such that the Major version increases,
        a new version of this module will likely need to be used.

        Returns:
            API version (:obj:`float`).

        Example:
            >>> client = SparclClient()
            >>> client.version
            12.0
        """

        if self.apiversion is None:
            response = requests.get(
                f"{self.apiurl}/version", timeout=self.timeout, cache=True
            )
            self.apiversion = float(response.content)
        return self.apiversion

    def find(
        self,
        outfields=None,
        *,
        constraints={},  # dict(fname) = [op, param, ...]
        #! exclude_unauth = True,  # Not implemented yet
        limit=500,
        sort=None,
        # count=False,
        # dataset_list=None,
        verbose=None,
    ):
        """Find records in the SPARCL database.

        Args:
            outfields (:obj:`list`, optional): List of fields to return.
                Only CORE fields may be passed to this parameter.
                Defaults to None, which will return only the sparcl_id
                and _dr fields.

            constraints (:obj:`dict`, optional): Key-Value pairs of
                constraints to place on the record selection. The Key
                part of the Key-Value pair is the field name and the
                Value part of the Key-Value pair is a list of values.
                Defaults to no constraints. This will return all records in the
                database subject to restrictions imposed by the ``limit``
                parameter.

            limit (:obj:`int`, optional): Maximum number of records to
                return. Defaults to 500.

            sort (:obj:`list`, optional): Comma separated list of fields
                to sort by. Defaults to None. (no sorting)

            verbose (:obj:`bool`, optional): Set to True for in-depth return
                statement. Defaults to False.

        Returns:
            :class:`~sparcl.Results.Found`: Contains header and records.

        Example:
            >>> client = SparclClient()
            >>> outs = ['sparcl_id', 'ra', 'dec']
            >>> cons = {'spectype': ['GALAXY'], 'redshift': [0.5, 0.9]}
            >>> found = client.find(outfields=outs, constraints=cons)
            >>> sorted(list(found.records[0].keys()))
            ['_dr', 'dec', 'ra', 'sparcl_id']
        """
        # dataset_list (:obj:`list`, optional): List of data sets from
        #     which to find records. Defaults to None, which
        #     will find records in all data sets hosted on the SPARC
        #     database.

        verbose = self.verbose if verbose is None else verbose

        # Let "outfields" default to ['id']; but fld may have been renamed
        if outfields is None:
            outfields = ["sparcl_id"]
        dataset_list = self.fields.all_drs
        #! self._validate_science_fields(outfields,
        #!                               dataset_list=dataset_list) # DLS-401
        dr = list(dataset_list)[0]
        if len(constraints) > 0:
            self._validate_science_fields(
                constraints.keys(), dataset_list=dataset_list
            )
            constraints = {
                self.fields._internal_name(k, dr): v
                for k, v in constraints.items()
            }
        uparams = dict(
            limit=limit,
            #! count='Y' if count else 'N'
        )
        if sort is not None:
            uparams["sort"] = sort
        qstr = urlencode(uparams)
        url = f"{self.apiurl}/find/?{qstr}"

        outfields = [self.fields._internal_name(s, dr) for s in outfields]
        search = [[k] + v for k, v in constraints.items()]
        sspec = dict(outfields=outfields, search=search)

        if verbose:
            print(f"url={url} sspec={sspec}")
        if self.show_curl:
            cmd = ut.curl_find_str(sspec, self.rooturl, qstr=qstr)
            print(cmd)

        auth = TokenAuth(self.token, self.token_expired) if self.token else None  # noqa: E501
        res = requests.post(url, json=sspec, auth=auth, timeout=self.timeout)

        if res.status_code != 200:
            if verbose and ("traceback" in res.json()):
                print(f'DBG: Server traceback=\n{res.json()["traceback"]}')
            raise ex.genSparclException(res, verbose=self.verbose)

        found = Found(res.json(), client=self)
        if verbose:
            print(f"Record key counts: {ut.count_values(found.records)}")
        return found

    def missing(
        self, uuid_list, *, dataset_list=None, countOnly=False, verbose=False
    ):
        """Return the subset of sparcl_ids in the given uuid_list that are
        NOT stored in the SPARCL database.

        Args:
            uuid_list (:obj:`list`): List of sparcl_ids.

            dataset_list (:obj:`list`, optional): List of data sets from
                which to find missing sparcl_ids. Defaults to None, meaning
                all data sets hosted on the SPARCL database.

            countOnly (:obj:`bool`, optional): Set to True to return only
                a count of the missing sparcl_ids from the uuid_list.
                Defaults to False.

            verbose (:obj:`bool`, optional): Set to True for in-depth return
                statement. Defaults to False.

        Returns:
            A list of the subset of sparcl_ids in the given uuid_list that
            are NOT stored in the SPARCL database.

        Example:
            >>> client = SparclClient()
            >>> ids = ['ddbb57ee-8e90-4a0d-823b-0f5d97028076',]
            >>> client.missing(ids)
            ['ddbb57ee-8e90-4a0d-823b-0f5d97028076']
        """

        if dataset_list is None:
            dataset_list = self.fields.all_drs
        assert isinstance(
            dataset_list, (list, set)
        ), f"DATASET_LIST must be a list. Found {dataset_list}"

        verbose = verbose or self.verbose
        uparams = dict(dataset_list=",".join(dataset_list))
        qstr = urlencode(uparams)
        url = f"{self.apiurl}/missing/?{qstr}"
        uuids = list(uuid_list)
        if verbose:
            print(f'Using url="{url}"')
        res = requests.post(url, json=uuids, timeout=self.timeout)

        res.raise_for_status()
        if res.status_code != 200:
            raise Exception(res)
        ret = res.json()
        if countOnly:
            return len(ret)
        return ret
        # END missing()

    def missing_specids(
        self, specid_list, *, dataset_list=None, countOnly=False, verbose=False
    ):
        """Return the subset of specids in the given specid_list that are
        NOT stored in the SPARCL database.

        Args:
            specid_list (:obj:`list`): List of specids.

            dataset_list (:obj:`list`, optional): List of data sets from
                which to find missing specids. Defaults to None, meaning
                all data sets hosted on the SPARCL database.

            countOnly (:obj:`bool`, optional): Set to True to return only
                a count of the missing specids from the specid_list.
                Defaults to False.

            verbose (:obj:`bool`, optional): Set to True for in-depth return
                statement. Defaults to False.

        Returns:
            A list of the subset of specids in the given specid_list that
            are NOT stored in the SPARCL database.

        Example:
            >>> client = SparclClient()
            >>> found = client.find(outfields=['specid'], limit=2)
            >>> specids = [f.specid for f in found.records]
            >>> client.missing_specids(specids + ['bad_id'])
            ['bad_id']
        """
        if dataset_list is None:
            dataset_list = self.fields.all_drs
        assert isinstance(
            dataset_list, (list, set)
        ), f"DATASET_LIST must be a list. Found {dataset_list}"

        verbose = verbose or self.verbose
        uparams = dict(dataset_list=",".join(dataset_list))
        qstr = urlencode(uparams)
        url = f"{self.apiurl}/missing_specids/?{qstr}"
        specids = list(specid_list)
        if verbose:
            print(f'Using url="{url}"')
        res = requests.post(url, json=specids, timeout=self.timeout)

        res.raise_for_status()
        if res.status_code != 200:
            raise Exception(res)
        ret = res.json()
        if countOnly:
            return len(ret)
        return ret
        # END missing_specids()

    # Include fields are Science (not internal) names. But the mapping
    # of Internal to Science name depends on DataSet.  Its possible
    # for a field (Science name) to be valid in one DataSet but not
    # another.  For the include_list to be valid, all fields must be
    # valid Science field names for all DS in given dataset_list.
    # (defaults to all DataSets ingested)
    def _validate_include(self, include_list, dataset_list):
        if not isinstance(include_list, (list, set)):
            msg = f"Bad INCLUDE_LIST. Must be list. Got {include_list}"
            raise ex.BadInclude(msg)

        avail_science = self.get_available_fields(dataset_list=dataset_list)
        inc_set = set(include_list)
        unknown = inc_set.difference(avail_science)
        if len(unknown) > 0:
            msg = (
                f'The INCLUDE list ({",".join(sorted(include_list))}) '
                f"contains invalid data field names "
                f'for Data Sets ({",".join(sorted(dataset_list))}). '
                f"Unknown fields are: "
                f'{", ".join(sorted(list(unknown)))}. '
                f"Available fields are: "
                f'{", ".join(sorted(avail_science))}.'
            )
            raise ex.BadInclude(msg)
        return True

    def retrieve(  # noqa: C901
        self,
        uuid_list,
        *,
        include="DEFAULT",
        dataset_list=None,
        limit=500,
        verbose=None,
    ):
        """Retrieve spectra records from the SPARCL database by list of
        sparcl_ids.

        Args:
            uuid_list (:obj:`list`): List of sparcl_ids.

            include (:obj:`list`, optional): List of field names to include
                in each record. Defaults to 'DEFAULT', which will return
                the fields tagged as 'default'.

            dataset_list (:obj:`list`, optional): List of data sets from
                which to retrieve spectra data. Defaults to None, meaning all
                data sets hosted on the SPARCL database.

            limit (:obj:`int`, optional): Maximum number of records to
                return. Defaults to 500. Maximum allowed is 24,000.

            verbose (:obj:`bool`, optional): Set to True for in-depth return
                statement. Defaults to False.

        Returns:
            :class:`~sparcl.Results.Retrieved`: Contains header and records.

        Example:
            >>> client = SparclClient()
            >>> ids = client.find(limit=1).ids
            >>> inc = ['sparcl_id', 'flux', 'wavelength', 'model']
            >>> ret = client.retrieve(uuid_list=ids, include=inc)
            >>> type(ret.records[0].wavelength)
            <class 'numpy.ndarray'>
        """

        # Variants for async, etc.
        #
        # From "performance testing" docstring
        #    svc (:obj:`str`, optional): Defaults to 'spectras'.
        #
        #    format (:obj:`str`, optional): Defaults to 'pkl'.
        #
        #
        #    chunk (:obj:`int`, optional): Size of chunks to break list into.
        #        Defaults to 500.
        #
        # These were keyword params:
        svc = "spectras"  # retrieve, spectras
        format = "pkl"  # 'json',
        chunk = 500

        orig_dataset_list = dataset_list
        if dataset_list is None:
            dataset_list = self.fields.all_drs
        assert isinstance(
            dataset_list, (list, set)
        ), f"DATASET_LIST must be a list. Found {dataset_list}"

        verbose = self.verbose if verbose is None else verbose

        if (include == DEFAULT) or (include is None) or include == []:
            include_list = self.get_default_fields(dataset_list=dataset_list)
        elif include == ALL:
            include_list = self.get_all_fields(dataset_list=dataset_list)
        else:
            include_list = include

        self._validate_include(include_list, dataset_list)

        req_num = min(len(uuid_list), (limit or len(uuid_list)))
        #! print(f'DBG: req_num = {req_num:,d}'
        #!       f'  len(uuid_list)={len(uuid_list):,d}'
        #!       f'  limit={limit}'
        #!       f'  MAX_NUM_RECORDS_RETRIEVED={MAX_NUM_RECORDS_RETRIEVED:,d}')
        if req_num > MAX_NUM_RECORDS_RETRIEVED:
            msg = (
                f"Too many records asked for with client.retrieve()."
                f"  {len(uuid_list):,d} IDs provided,"
                f"  limit={limit}."
                f"  But the maximum allowed is"
                f" {MAX_NUM_RECORDS_RETRIEVED:,d}."
            )
            raise ex.TooManyRecords(msg)

        com_include = self._common_internal(
            science_fields=include_list, dataset_list=dataset_list
        )
        uparams = {
            "include": ",".join(com_include),
            # limit=limit,  # altered uuid_list to reflect limit
            #! "chunk_len": chunk,
            "format": format,
            #! "1thread": "yes",  # @@@ 7.3.2023
            #!"dataset_list": ",".join(dataset_list),
        }
        # Do not put dataset_list in server call if it wasn't in client call.
        # Else will interfer with defaulting behavior of AUTH (DLS-504).
        if orig_dataset_list is not None:
            uparams["dataset_list"] = ",".join(dataset_list)

        qstr = urlencode(uparams)

        #!url = f'{self.apiurl}/retrieve/?{qstr}'
        url = f"{self.apiurl}/{svc}/?{qstr}"
        if verbose:
            print(f'Using url="{url}"')
            ut.tic()

        ids = list(uuid_list) if limit is None else list(uuid_list)[:limit]
        if self.show_curl:
            cmd = ut.curl_retrieve_str(ids, self.rooturl, svc=svc, qstr=qstr)
            print(cmd)

        try:
            auth = TokenAuth(self.token, self.token_expired) if self.token else None  # noqa: E501
            res = requests.post(url, json=ids, auth=auth, timeout=self.timeout)
        except requests.exceptions.ConnectTimeout as reCT:
            raise ex.UnknownSparcl(f"ConnectTimeout: {reCT}")
        except requests.exceptions.ReadTimeout as reRT:
            msg = (
                f'Try increasing the value of the "read_timeout" parameter'
                f' to "SparclClient()".'
                f" The current values is: {self.r_timeout} (seconds)"
                f"{reRT}"
            )
            raise ex.ReadTimeout(msg) from None
        except requests.exceptions.ConnectionError as reCE:
            raise ex.UnknownSparcl(f"ConnectionError: {reCE}")
        except requests.exceptions.TooManyRedirects as reTMR:
            raise ex.UnknownSparcl(f"TooManyRedirects: {reTMR}")
        except requests.exceptions.HTTPError as reHTTP:
            raise ex.UnknownSparcl(f"HTTPError: {reHTTP}")
        except requests.exceptions.URLRequired as reUR:
            raise ex.UnknownSparcl(f"URLRequired: {reUR}")
        except requests.exceptions.RequestException as reRE:
            raise ex.UnknownSparcl(f"RequestException: {reRE}")
        except Exception as err:  # fall through
            raise ex.UnknownSparcl(err)

        if verbose:
            elapsed = ut.toc()
            print(f"Got response to post in {elapsed} seconds")
        if res.status_code != 200:
            if verbose:
                print(f"DBG: Server response=\n{res.text}")
            # @@@ FAILS on invalid JSON. Maybe not json at all !!!
            if verbose and ("traceback" in res.json()):
                print(f'DBG: Server traceback=\n{res.json()["traceback"]}')
            raise ex.genSparclException(res, verbose=verbose)

        if format == "json":
            results = res.json()
        elif format == "pkl":
            # Read chunked binary file (representing pickle file) from
            # server response. Load pickle into python data structure.
            # Python structure is list of records where first element
            # is a header.
            with tempfile.TemporaryFile(mode="w+b") as fp:
                for idx, chunk in enumerate(res.iter_content(chunk_size=None)):
                    fp.write(chunk)
                # Position to start of file for pickle reading (load)
                fp.seek(0)
                results = pickle.load(fp)
        else:
            results = res.json()

        meta = results[0]
        if verbose:
            count = len(results) - 1
            print(
                f"Got {count} spectra in "
                f"{elapsed:.2f} seconds ({count/elapsed:.0f} "
                "spectra/sec)"
            )
            print(f'{meta["status"]}')

        # Format/consolodate the server messages to one message with
        # the count of missing files
        if len(meta["status"].get("warnings", [])) > 0:
            warnings = meta["status"].get("warnings")
            if verbose:
                print(f"There are {len(warnings)} warnings")
            missingcount = 0
            missing_message = re.sub(r' [0-9]+ ', ' %s ', warnings[0])
            for i in warnings:
                matches = re.match(r'.* ([0-9]+) out of the ([0-9]+).*', i)
                if matches:
                    missingcount += int(matches.groups()[0])

            # using old style substitution to avoid issue with the {} in the message  # noqa: E501
            warning_message = missing_message % (missingcount, req_num, missingcount)   # noqa: E501
            warn(warning_message, stacklevel=2)

        return Retrieved(results, client=self)

    def retrieve_by_specid(
        self,
        specid_list,
        *,
        svc="spectras",  # 'retrieve',
        format="pkl",  # 'json',
        include="DEFAULT",
        dataset_list=None,
        limit=500,
        verbose=False,
    ):
        """Retrieve spectra records from the SPARCL database by list of
        specids.

        Args:
            specid_list (:obj:`list`): List of specids.

            include (:obj:`list`, optional): List of field names to include
                in each record. Defaults to 'DEFAULT', which will return
                the fields tagged as 'default'.

            dataset_list (:obj:`list`, optional): List of data sets from
                which to retrieve spectra data. Defaults to None, meaning all
                data sets hosted on the SPARCL database.

            limit (:obj:`int`, optional): Maximum number of records to
                return. Defaults to 500. Maximum allowed is 24,000.

            verbose (:obj:`bool`, optional): Set to True for in-depth return
                statement. Defaults to False.

        Returns:
            :class:`~sparcl.Results.Retrieved`: Contains header and records.

        Example:
            >>> client = SparclClient()
            >>> sids = [5840097619402313728, -8985592895187431424]
            >>> inc = ['specid', 'flux', 'wavelength', 'model']
            >>> ret = client.retrieve_by_specid(specid_list=sids, include=inc)
            >>> len(ret.records[0].wavelength)
            4617

        """
        #!specid_list = list(specid_list)
        assert isinstance(specid_list, list), (
            f'The "specid_list" parameter must be a python list. '
            f"You used a value of type {type(specid_list)}."
        )
        assert (
            len(specid_list) > 0
        ), f'The "specid_list" parameter value must be a non-empty list'
        assert isinstance(specid_list[0], int), (
            f'The "specid_list" parameter must be a python list of INTEGERS. '
            f"You used an element value of type {type(specid_list[0])}."
        )

        if dataset_list is None:
            constraints = {"specid": specid_list}
        else:
            constraints = {"specid": specid_list, "data_release": dataset_list}

        # Science Field Name for uuid.
        dr = list(self.fields.all_drs)[0]
        idfld = self.fields._science_name("sparcl_id", dr)

        found = self.find([idfld], constraints=constraints, limit=limit)
        if verbose:
            print(f"Found {found.count} matches.")
        res = self.retrieve(
            found.ids,
            #! svc=svc,
            #! format=format,
            include=include,
            dataset_list=dataset_list,
            limit=limit,
            verbose=verbose,
        )
        if verbose:
            print(f"Got {res.count} records.")
        return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# Python Standard Library
import configparser
import os.path


class Conf:
    """
    Configuration parameters for `ada_client`.
    """

    def __init__(self, conf_file=None):
        config = configparser.ConfigParser()
        conf_files = ["~/sparc.ini", "sparcl/sparc.ini"]
        if conf_file is None:
            for cf in conf_files:
                if os.path.exists(os.path.expanduser(cf)):
                    config.read(os.path.expanduser(cf))

        if "ada.server" not in config:
            raise Exception(
                f"Could not find conf file in any of: "
                f'{(",").join(conf_files)} '
                f"Create one and try again."
            )

        self.config = config

    @property
    def server_baseurl(self):
        return self.config["sparc.server"]["ServerBaseUrl"]

    @property
    def server_timeout(self):
        return self.config["sparc.server"]["ServerTimout"]



"""
A client for getting spectra and meta-data from NOIRLab.
"""

# List of packages to import when "from sparcl import *" is used
__all__ = ["client", "align_records"]


# See semantic versioning

# BUT PyPi requires honoring versions like this:
# https://packaging.python.org/specifications/core-metadata/
# https://www.python.org/dev/peps/pep-0440/
#
# '0.3.0-alpha3.23' is an invalid value for Version.
#  Error: Start and end with a letter or numeral containing only ASCII
#  numeric and '.', '_' and '-'.
#
# https://semver.org/ yields possible versions that violate PEP-0440

# __version__ = '0.3.21'
# __version__ = '0.1a3.dev22'
# __version__ = '0.3.0-alpha3.23'
# __version__ = '0.3.22'

# must mach: [N!]N(.N)*[{a|b|rc}N][.postN][.devN]
# Example of a correct version string: '0.4.0a3.dev35'
# __version__ = '1.1'
# __version__ = '1.2.0b4'
# __version__ = '1.2.0'  # Release
# __version__ = "1.2.1b3"
# __version__ = "1.2.1"
# FIRST uncommented value will be used! (so only leave one uncommented)
#__version__ = "1.2.2"
#__version__ = "1.2.3"
#__version__ = "1.2.4"
#__version__ = "1.2.5"
__version__ = "1.2.6b7"



"""Containers for results from SPARCL Server.
These include results of client.retrieve() client.find().
"""

from collections import UserList

#!import copy
from sparcl.utils import _AttrDict

# from sparcl.gather_2d import bin_spectra_records
import sparcl.exceptions as ex
from warnings import warn
import re

class Results(UserList):
    def __init__(self, dict_list, client=None):
        super().__init__(dict_list)
        self.hdr = dict_list[0]
        self.recs = dict_list[1:]
        self.client = client
        self.fields = client.fields if client else []
        self.to_science_fields()

        # HACK 12/14/2023 -sp- to fix UUID problem presumably
        # produced on stack version upgrade (to Django 4.2, postgres 13+)
        # Done per AB for expediency since real solution will be easier
        # after field-renaming is removed.
        for rec in self.recs:
            if "sparcl_id" in rec:
                rec["sparcl_id"] = str(rec["sparcl_id"])
        # END __init__()

    # https://docs.python.org/3/library/collections.html#collections.deque.clear
    def clear(self):
        """Delete the contents of this collection."""
        super().clear()
        self.hdr = {}
        self.recs = []

    @property
    def info(self):
        """Info about this collection.
        e.g. Warnings, parameters used to get the collection, etc."""
        # Consolodate "Successfully found...." messages
        cln_hdr = self.hdr.copy()
        count = 0
        success_msg = ""
        info = []
        for msg in cln_hdr['status']['info']:
            if "Successfully" in msg:
                matches = re.search('Successfully found ([0-9]*)', msg)
                count += int(matches.groups()[0]) if len(matches.groups()) > 0 else 0  # noqa: E501
                success_msg = msg
            else:
                info.append(msg)

        if count > 0:
            msg = re.sub(r'[0-9]+', str(count), success_msg, count=1)
            info.append(msg)

        cln_hdr['status']['info'] = info
        return cln_hdr

    @property
    def count(self):
        """Number of records in this collection."""
        return len(self.recs)

    @property
    def records(self):
        """Records in this collection. Each record is a dictionary."""
        return self.recs

    def json(self):
        return self.data

    # Convert Internal field names to Science field names.
    # SIDE-EFFECT: modifies self.recs
    def to_science_fields(self):  # from_orig
        newrecs = list()
        for rec in self.recs:
            newrec = dict()
            dr = rec["_dr"]
            keep = True
            for orig in rec.keys():
                if orig == "_dr":
                    # keep DR around unchanged. We need it to rename back
                    # to Internal Field Names later.
                    newrec[orig] = rec[orig]
                else:
                    new = self.fields._science_name(orig, dr)
                    if new is None:
                        keep = False  # We don't have name mapping, toss rec
                    newrec[new] = rec[orig]
            if keep:
                newrecs.append(_AttrDict(newrec))
        self.recs = newrecs

    # Convert Science field names to Internal field names.
    def to_internal_fields(self):
        for rec in self.recs:
            dr = rec.get("_dr")
            for new in rec.keys():
                if new == "_dr":
                    # keep DR around unchanged. We need it to rename back
                    # to Internal Field Names later.
                    continue
                new = self.fields._internal_name(new, dr)
                rec[new] = rec.pop(new)

    def science_to_internal_fields(self):
        newrecs = list()
        for rec in self.recs:
            newrec = dict()
            dr = rec["_dr"]
            keep = True
            for sci_name in rec.keys():
                if sci_name == "_dr":
                    # keep DR around unchanged. We need it to rename back
                    # to Internal Field Names later.
                    newrec[sci_name] = rec[sci_name]
                else:
                    new = self.fields._internal_name(sci_name, dr)
                    if new is None:
                        keep = False
                    newrec[new] = rec[sci_name]
            if keep:
                newrecs.append(_AttrDict(newrec))
        self.recs = newrecs
        return self.recs

    def reorder(self, ids_og):
        """
        Reorder the retrieved records to be in the same
        order as the original IDs passed to client.retrieve().

        Args:
            ids_og (:obj:`list`): List of sparcl_ids or specIDs.

        Returns:
            reordered (:class:`~sparcl.Results.Retrieved`): Contains header and
                                                            reordered records.
            # none_idx (:obj:`list`): List of indices where record is None.

        """
        if len(ids_og) <= 0:
            msg = (
                f"The list of IDs passed to the reorder method "
                f"does not contain any sparcl_ids or specIDs."
            )
            raise ex.NoIDs(msg)
        elif len(self.recs) <= 0:
            msg = (
                "The retrieved or found results did not "
                "contain any records."
            )
            raise ex.NoRecords(msg)
        else:
            # Transform science fields to internal fields
            new_recs = self.science_to_internal_fields()
            # Get the ids or specids from retrieved records
            if type(ids_og[0]) is str:
                ids_re = [f["sparcl_id"] for f in new_recs]
            elif type(ids_og[0]) is int:
                ids_re = [f["specid"] for f in new_recs]
            # Enumerate the original ids
            dict_og = {x: i for i, x in enumerate(ids_og)}
            # Enumerate the retrieved ids
            dict_re = {x: i for i, x in enumerate(ids_re)}
            # Get the indices of the original ids. Set to None if not found
            idx = [dict_re.get(key, None) for key in dict_og.keys()]
            # Get the indices of None values
            none_idx = [i for i, v in enumerate(idx) if v is None]
            # Reorder the retrieved records
            reordered = [self.recs[i] for i in idx if i is not None]
            # Insert dummy record(s) if applicable
            dummy_record = "{'id': None, 'specid': None, '_dr': 'SDSS-DR16'}"
            for i in none_idx:
                reordered.insert(
                    i, {"sparcl_id": None, "specid": None, "_dr": "SDSS-DR16"}
                )
            reordered.insert(0, self.hdr)
            meta = reordered[0]
            if len(none_idx) > 0:
                msg = (
                    f"{len(none_idx)} sparcl_ids or specIDs were "
                    f"not found in "
                    f'the database. Use "client.missing()" '
                    f"to get a list of the unavailable IDs. "
                    f"To maintain correct reordering, a dummy "
                    f"record has been placed at the indices "
                    f"where no record was found. Those "
                    f"indices are: {none_idx}. The dummy "
                    f"record will appear as follows: "
                    f"{dummy_record}. "
                )
                meta["status"].update({"warnings": [msg]})
                warn(msg, stacklevel=2)
        return Results(reordered, client=self.client)


# For results of retrieve()
class Retrieved(Results):
    """Holds spectra records (and header)."""

    def __init__(self, dict_list, client=None):
        super().__init__(dict_list, client=client)

    def __repr__(self):
        return f"Retrieved Results: {len(self.recs)} records"


#!    def bin_spectra(self):
#!        """Align flux from all records by common wavelength bin.
#!
#!        A value of nan is used where a record does not contain a flux
#!        value for a specific bin.
#!
#!        Returns:
#!           flux: 2d numpy array with shape (numRecords, numWavelengthBins)
#!                 Flux value for each record, each bin
#!           wavs: 1d numpy array with shape (numWavelengthBins)
#!                 Wavelength values for each bin
#!
#!        Example:
#!            >>> client = sparcl.client.SparclClient()
#!            >>> found = client.find(
#!                            constraints={"data_release": ['BOSS-DR16']},
#!                            limit=10)
#!            >>> got = client.retrieve(found.ids)
#!            >>> flux2d,wavs = got.bin_spectra()
#!
#!        """
#!        flux2d, wavs = bin_spectra_records(self.recs)
#!        return flux2d, wavs


class Found(Results):
    """Holds metadata records (and header)."""

    def __init__(self, dict_list, client=None):
        super().__init__(dict_list, client=client)

    def __repr__(self):
        return f"Find Results: {len(self.recs)} records"

    @property
    def ids(self):
        """List of unique identifiers of matched records."""
        #! dr = list(self.fields.all_drs)[0]

        return [d.get("sparcl_id") for d in self.recs]



# NOT INTENDED FOR PUBLIC USE!
#
# See:
#   https://spectres.readthedocs.io/en/latest/
import math
import spectres
import numpy as np

# Local
import sparcl.client


# Per paper, should be able to pass all flux in one call to spectres
# https://arxiv.org/pdf/1705.05165.pdf
# Perhaps users would rather the bins uniform (1,5,20 Angstroms?)
def _resample_flux(records, wavstep=1):
    smallest = math.floor(min([min(r.wavelength) for r in records]))
    largest = math.ceil(max([max(r.wavelength) for r in records]))

    #!wrange = largest - smallest
    # new_wavs = np.fro<mfunction(lambda i: i + smallest, (wrange,), dtype=int)
    # flux_2d = np.ones([len(records), wrange])

    new_wavs = np.array(range(smallest, largest + 1, wavstep))
    flux_2d = np.full([len(records), len(new_wavs)], None, dtype=float)

    for idx, rec in enumerate(records):
        flux_2d[idx] = spectres.spectres(
            new_wavs, rec.wavelength, rec.flux, verbose=False
        )
    return flux_2d, new_wavs


def _tt0(numrecs=20):
    client = sparcl.client.SparclClient()
    found = client.find(
        constraints=dict(data_release=["BOSS-DR16"]), limit=numrecs
    )
    got = client.retrieve(found.ids)
    flux_2d, new_wavs = _resample_flux(got.records)
    return flux_2d, new_wavs



"""Get Field names associated with various SPARCL conditions.
"""

# Python Standard Library
from collections import defaultdict

# External Packages
import requests


def validate_fields(datafields):
    # datafields is simply:
    #   DataField.objects.all().values(*atts)

    drs = set([df["data_release"] for df in datafields])
    core = {
        df["origdp"]: df["newdp"] for df in datafields if df["storage"] == "C"
    }

    o2n = {
        dr: {
            df["origdp"]: df["newdp"]
            for df in datafields
            if df["data_release"] == dr
        }
        for dr in drs
    }

    for dr, df in o2n.items():
        #  1-1 mapping origdp <-> newdp across all DR
        if len(set(df.values())) != len(df):
            msg = (
                f"Data Release={dr} does not have a one-to-one mapping "
                f"between Original and Science field names."
            )
            raise Exception(msg)

        acore = defaultdict(list)  # ambiguous core fields(more than one value)
        for k in core.keys():
            if df.get(k) != core.get(k):
                acore[k].append(df.get(k))
        if len(acore) > 0:
            msg = (
                f"DataFields do not have the same "
                f"Science field name for core values across all Data Sets. "
                f"{dict(acore)}"
            )
            raise Exception(msg)

    return True


class Fields:  # Derived from a single query
    """Lookup of Field Names"""

    def __init__(self, apiurl):
        # [rec, ...]
        # where rec is dict containing keys:
        # 'data_release', 'origdp', 'newdp', 'storage', 'default', 'all'
        datafields = requests.get(f"{apiurl}/datafields/").json()

        validate_fields(datafields)

        dr_list = set(df["data_release"] for df in datafields)

        self.datafields = datafields
        # o2n[DR][InternalName] => ScienceName
        self.o2n = {
            dr: {
                df["origdp"]: df["newdp"]
                for df in datafields
                if df["data_release"] == dr
            }
            for dr in dr_list
        }
        # n2o[DR][ScienceName] => InternalName
        self.n2o = {
            dr: {
                df["newdp"]: df["origdp"]
                for df in datafields
                if df["data_release"] == dr
            }
            for dr in dr_list
        }
        self.all_drs = dr_list
        self.all_fields = set([df["newdp"] for df in datafields])
        self.datafields = datafields

        # Per DataRelease: get Storage, Default, All for each (user) fieldname
        # dr_attrs[DR][newdp] => dict[storage,default,all]
        self.attrs = {
            dr: {
                df["newdp"]: {
                    "storage": df["storage"],
                    "default": df["default"],
                    "all": df["all"],
                }
                for df in datafields
                if df["data_release"] == dr
            }
            for dr in dr_list
        }

    @property
    def all_datasets(self):
        return self.all_drs

    def _science_name(self, internal_name, dataset):
        return self.o2n[dataset].get(internal_name)

    def _internal_name(self, science_name, dataset):
        #!return self.n2o[dataset][science_name]
        return self.n2o[dataset].get(science_name)

    def filter_fields(self, attr, dataset_list):
        fields = set()
        for dr in dataset_list:
            for k, v in self.attrs[dr].items():
                if v.get(attr):
                    fields.add(k)
        return fields

    def default_retrieve_fields(self, dataset_list=None):
        if dataset_list is None:
            dataset_list = self.all_drs
        return self.filter_fields("default", dataset_list)

    def all_retrieve_fields(self, dataset_list=None):
        if dataset_list is None:
            dataset_list = self.all_drs
        return self.filter_fields("all", dataset_list)

    def common(self, dataset_list=None):
        """Fields common to DATASET_LIST (or All datasets if None)"""
        if dataset_list is None:
            dataset_list = self.all_drs
        return sorted(
            set.intersection(
                *[set(self.n2o[dr].keys()) for dr in dataset_list]
            )
        )

    def common_internal(self, dataset_list=None):
        """Fields common to DATASET_LIST (or All datasets if None)"""
        if dataset_list is None:
            dataset_list = self.all_drs
        return set.intersection(
            *[set(self.o2n[dr].keys()) for dr in dataset_list]
        )

    # There is probably an algorithm to partition ELEMENTS into
    # the _minumum_ number of SETS such that the union of all SETS
    # contains all ELEMENTS. For now, parition by Data Set (when used).
    def field_partitions(self, fields):
        """Partition FIELDS into the DataSets that contain them"""
        dr_fields = defaultdict(list)
        for field in fields:
            for dr in self.all_drs:
                if field in self.n2o[dr]:
                    dr_fields[dr].append(field)
        return dict(dr_fields)



# For running client against PAT host
# See also: expected_dev1.py   (SP DEV)

all_fields = [
    "data_release",
    "datasetgroup",
    "dateobs",
    "dateobs_center",
    "dec",
    "exptime",
    "flux",
    "instrument",
    "ivar",
    "mask",
    "model",
    "ra",
    "redshift",
    "redshift_err",
    "redshift_warning",
    "site",
    "sparcl_id",
    "specid",
    "specprimary",
    "spectype",
    "survey",
    "targetid",
    "telescope",
    "wave_sigma",
    "wavelength",
    "wavemax",
    "wavemin",
]


default_fields = ["dec", "flux", "ra", "sparcl_id", "specid", "wavelength"]

retrieve_0 = ["_dr", "flux", "sparcl_id", "specid"]

retrieve_0b = ["_dr", "dec", "flux", "ra", "sparcl_id", "specid", "wavelength"]

retrieve_5 = 2

# OLD as of August 23, 2024
#find_0 = [
#    {
#        "_dr": "BOSS-DR16",
#        "data_release": "BOSS-DR16",
#        "specid": -6444532452352045056,
#    }
#]

find_0 = [
    {
        '_dr': 'BOSS-DR16',
        'data_release': 'BOSS-DR16',
        'telescope': 'sloan25m'
    }
]

find_1 = [
    {
        "_dr": "SDSS-DR16",
        "data_release": "SDSS-DR16",
        "specid": 1506454396622366720,
    }
]

find_2 = 936894  # PAT

# OLD as of August 23, 2024
#find_3 = [
#    {"_dr": "BOSS-DR16", "data_release": "BOSS-DR16"},
#    {"_dr": "DESI-EDR", "data_release": "DESI-EDR"},
#    {"_dr": "SDSS-DR16", "data_release": "SDSS-DR16"},
#]

find_4 = 36

find_5a = [
    {"_dr": "BOSS-DR16", "data_release": "BOSS-DR16", "mjd": 55689},
    {"_dr": "SDSS-DR16", "data_release": "SDSS-DR16", "mjd": 54763},
]

find_5d = []

authorized_1 = {
    "Loggedin_As": "test_user_1@noirlab.edu",
    "Authorized_Datasets": {
        "BOSS-DR16",
        "BOSS-DR17",
        "DESI-DR1",
        "DESI-EDR",
        "SDSS-DR16",
        "SDSS-DR17",
        "SDSS-DR17-test",
    },
}

authorized_2 = {
    "Loggedin_As": "test_user_2@noirlab.edu",
    "Authorized_Datasets": {"BOSS-DR16", "DESI-DR1", "DESI-EDR", "SDSS-DR16"},
}

authorized_3 = {
    "Loggedin_As": "Anonymous",
    "Authorized_Datasets": {"BOSS-DR16", "DESI-DR1", "DESI-EDR", "SDSS-DR16"},
}

# Private and Public
pub_1 = ["BOSS-DR16"]
pub_all = ["BOSS-DR16", "DESI-DR1", "DESI-EDR", "SDSS-DR16"]
priv = ["SDSS-DR17-test", "BOSS-DR17", "SDSS-DR17"]
all_all = pub_all + priv
all_all.sort()
privpub1 = [pub_all[0], priv[0]]
unauth = "test_user_2@noirlab.edu"
msg = "Find Results: 1000 records"
# OLD as of July 9, 2025
#auth_find_1 = privpub1
#auth_find_2 = ['DESI-DR1', 'DESI-EDR']
auth_find_1 = auth_find_2 = auth_find_4 = auth_find_6 = msg
auth_find_3 = auth_retrieve_3 = (f"[DSDENIED] uname='{unauth}' is declined "
                                 f"access to datasets={[priv[0]]}; "
                                 f"drs_requested={privpub1} "
                                 f"my_auth={pub_all}")
# OLD as of July 9, 2025
#auth_find_4 = auth_find_6 = pub_all
auth_find_5 = auth_retrieve_6 = ("[DSDENIED] uname='ANONYMOUS' is declined "
                                 f"access to datasets={[priv[0]]}; "
                                 f"drs_requested={privpub1} "
                                 f"my_auth={pub_all}")
auth_retrieve_1 = auth_retrieve_2 = privpub1
auth_retrieve_4 = auth_retrieve_5 = auth_retrieve_7 = auth_retrieve_8 = pub_1
